# ruff: noqa

# TODO: Might need to add Rich to pyinstaller part
import logging
from logging.config import dictConfig

from magnus import defaults

dictConfig(defaults.LOGGING_CONFIG)
logger = logging.getLogger(defaults.LOGGER_NAME)

from magnus.interaction import (
    end_interactive_session,
    get_experiment_tracker_context,
    get_from_catalog,
    get_object,
    get_parameter,
    get_run_id,
    get_run_log,
    get_secret,
    put_in_catalog,
    put_object,
    start_interactive_session,
    set_parameter,
    track_this,
)  # noqa
from magnus.sdk import Stub, Pipeline, Task, Parallel, Map, Catalog, Success, Fail  # noqa


# TODO: Think of model registry as a central place to store models.
# TODO: Implement Sagemaker pipelines as a executor.


# TODO: Think of way of generating dag hash without executor configuration
